@@ -11,13 +11,13 @@ import z3c.form.button
 import z3c.form.field
 import z3c.form.group
 import z3c.formui.form
-import zope.app.publication.browser
 import zope.component
 import zope.interface
 import zope.security
 import zope.security.interfaces
 import zope.traversing.api
 import zope.traversing.browser.interfaces
+import zope.traversing.publicationtraverse
 
 
 class display_title(object):
@@ -45,16 +45,41 @@ def all_(*constraints):
     return lambda form: all(constraint(form) for constraint in constraints)
 
 
-class BaseAddForm(z3c.formui.form.AddForm):
-    """Simple base add form."""
+def get_entities_util():
+    """Return the user entities utility."""
+    return zope.component.getUtility(icemac.addressbook.interfaces.IEntities)
+
+
+class BaseForm(object):
+    """Basis for all forms."""
 
     interface = None # interface for form
-    class_ = None # create object from this class
-    next_url = None # target after creation, one of ('object', 'parent')
 
     @property
     def fields(self):
-        return z3c.form.field.Fields(self.interface)
+        fields = get_entities_util().getEntity(self.interface)
+        return z3c.form.field.Fields(*fields.getFieldValuesInOrder())
+
+    def url(self, obj):
+        adapter = zope.component.getMultiAdapter(
+            (obj, self.request),
+            zope.traversing.browser.interfaces.IAbsoluteURL)
+        return adapter()
+
+
+class BaseAddForm(BaseForm, z3c.formui.form.AddForm):
+    """Simple base add form."""
+
+    class_ = None # create object from this class
+    next_url = None # target after creation, one of ('object', 'parent')
+
+
+    def createAndAdd(self, data):
+        # overwriting method as otherwise object created event would
+        # get send twice
+        obj = self.create(data)
+        self.add(obj)
+        return obj
 
     def create(self, data):
         return create(self, self.class_, data)
@@ -84,40 +109,38 @@ class BaseAddForm(z3c.formui.form.AddForm):
             context = self.context
         else:
             raise ValueError("Don't know how to handle next_url %r.")
-        return zope.component.getMultiAdapter(
-            (context, self.request),
-            zope.traversing.browser.interfaces.IAbsoluteURL)()
+        return self.url(context)
 
 
-class BaseEditForm(z3c.formui.form.EditForm):
+class BaseEditForm(BaseForm, z3c.formui.form.EditForm):
     """Base Edit form."""
 
-    next_url = None # target ojbect after edit, one of ('object', 'parent')
+    next_url = None # target object after edit, one of ('object', 'parent')
     next_view = None # target view after edit (None for default view)
-    interface = None # interface for form
-
-    @property
-    def fields(self):
-        return z3c.form.field.Fields(self.interface)
 
     def render(self):
+        if self.request.response.getStatus() in (302, 303, 304):
+            # redirecting
+            return ''
         if self.status in (self.successMessage, self.noChangesMessage):
             self.redirect_to_next_url()
         else:
             return super(BaseEditForm, self).render()
 
-    def redirect_to_next_url(self):
-        if self.next_url == 'object':
+    def redirect_to_next_url(self, next_url=None, next_view=None):
+        if next_url is None:
+            next_url = self.next_url
+        if next_url == 'object':
             target = self.context
-        elif self.next_url == 'parent':
+        elif next_url == 'parent':
             target = zope.traversing.api.getParent(self.context)
         else:
-            raise ValueError('next_url %r unknown' % self.next_url)
-        target_url = zope.component.getMultiAdapter(
-            (target, self.request),
-            zope.traversing.browser.interfaces.IAbsoluteURL)()
-        if self.next_view:
-            target_url += '/%s' % self.next_view
+            raise ValueError('next_url %r unknown' % next_url)
+        target_url = self.url(target)
+        if next_view is None:
+            next_view = self.next_view
+        if next_view:
+            target_url += '/%s' % next_view
         self.request.response.redirect(target_url)
 
     def applyChanges(self, data):
@@ -147,14 +170,15 @@ class BaseDeleteForm(BaseEditForm):
 
     label = _(u'Do you really want to delete this entry?')
     interface = None
-    field_names = () # tuple of field names for display or empty for all
+    field_names = () # tuple of field names for display; empty for all
+    next_view_after_delete = None # when None, use same view as self.next_view
 
     mode = z3c.form.interfaces.DISPLAY_MODE
     next_url = 'object'
 
     @property
     def fields(self):
-        fields = z3c.form.field.Fields(self.interface)
+        fields = super(BaseDeleteForm, self).fields
         if self.field_names:
             fields = fields.select(*self.field_names)
         return fields
@@ -163,12 +187,10 @@ class BaseDeleteForm(BaseEditForm):
     def handleCancel(self, action):
         self.status = self.noChangesMessage
 
-    @z3c.form.button.buttonAndHandler(_(u'Yes, delete it'), name='delete')
+    @z3c.form.button.buttonAndHandler(_(u'Yes, delete'), name='delete')
     def handleDelete(self, action):
-        self.next_url = 'parent'
-        self.redirect_to_next_url()
+        self.redirect_to_next_url('parent', self.next_view_after_delete)
         self._do_delete()
-        return ''
 
     def _do_delete(self):
         name = zope.traversing.api.getName(self.context)
@@ -184,14 +206,16 @@ class PrefixGroup(z3c.form.group.Group):
 
     @property
     def fields(self):
-        return z3c.form.field.Fields(self.interface, prefix=self.prefix)
+        fields = get_entities_util().getEntity(self.interface)
+        return z3c.form.field.Fields(
+            *fields.getFieldValuesInOrder(), **dict(prefix=self.prefix))
 
 
 def can_access(uri_part):
     """Create a button condition function to test whether the user can
     access the URL context/@@absolute_url/<uri_part>."""
     def can_access_form(form):
-        traverser = zope.app.publication.browser.PublicationTraverser()
+        traverser = zope.traversing.publicationtraverse.PublicationTraverser()
         try:
             view = traverser.traverseRelativeURL(
                 form.request, form.context, uri_part)
